/*----------------------------------------------------------------------
  File    : cm4seqs.h
  Contents: prefix tree management for closed and maximal sequences
  Author  : Christian Borgelt
  History : 2014.07.16 file created
----------------------------------------------------------------------*/
#ifndef __CM4SEQS__
#define __CM4SEQS__
#include "memsys.h"
#include "tract.h"
#include "report.h"

#ifndef INFINITY
#include <float.h>
#define INFINITY    (DBL_MAX+DBL_MAX)
#endif                          /* in case C99 is not supported */

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
/* --- target pattern types --- */
#define CMS_PERMUT   0x0100      /* partial permutation (no reps.) */
#define CMS_SEQUENCE 0x0200      /* sequence (repetitions possible) */
#define CMS_CLOSED   ISR_CLOSED  /* closed frequent patterns */
#define CMS_MAXIMAL  ISR_MAXIMAL /* maximal frequent patterns */

#define CMX_PERMUT   0x0100      /* partial permutation (no reps.) */
#define CMX_SEQUENCE 0x0200      /* sequence (repetitions possible) */
#define CMX_CLOSED   ISR_CLOSED  /* closed frequent patterns */
#define CMX_MAXIMAL  ISR_MAXIMAL /* maximal frequent patterns */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct cmsnode {        /* --- c/m prefix tree node --- */
  RSUPP          supp;          /* support of represented sequence */
  ITEM           item;          /* associated item (last item) */
  ITEM           height;        /* height of subtree rooted here */
  struct cmsnode *sibling;      /* successor node in sibling list */
  struct cmsnode *children;     /* list of child nodes */
} CMSNODE;                      /* (c/m prefix tree node) */

typedef struct {                /* --- c/m prefix tree --- */
  MEMSYS   *mem;                /* memory management system */
  int      target;              /* target pattern type */
  int      dir;                 /* direction of item order */
  ITEM     size;                /* number of items / array size */
  ITEM     item;                /* associated prefix item */
  ISREPORT *rep;                /* item set reporter for reporting */
  CMSNODE  root;                /* root node of the tree */
  ITEM     pos[1];              /* positions of items in sequence */
} CMSTREE;                      /* (c/m prefix tree) */

typedef struct {                /* --- c/m filter for sequences --- */
  int      target;              /* target pattern type */
  int      dir;                 /* direction of item order */
  ITEM     size;                /* number of items */
  ITEM     cnt;                 /* current number of prefix trees */
  CMSTREE  *trees[1];           /* conditional prefix trees */
} CM4SEQS;                      /* (closed/maximal filter for seqs.) */

/*----------------------------------------------------------------------
  Prefix Tree Functions
----------------------------------------------------------------------*/
extern CMSTREE* cms_create  (ITEM size, int target, int dir);
extern CMSTREE* cms_createx (MEMSYS *mem, ITEM size,
                             int target, int dir);
extern void     cms_clear   (CMSTREE *cms);
extern void     cms_delete  (CMSTREE *cms, int delms);
extern MEMSYS*  cms_memsys  (CMSTREE *cms);
extern ITEM     cms_cnt     (CMSTREE *cms);
extern int      cms_target  (CMSTREE *cms);
extern int      cms_dir     (CMSTREE *cms);
extern RSUPP    cms_supp    (CMSTREE *cms);
extern size_t   cms_nodecnt (CMSTREE *pat);
extern size_t   cms_nodemax (CMSTREE *pat);

extern int      cms_add     (CMSTREE *cms, const ITEM *items, ITEM n,
                             RSUPP supp);
extern RSUPP    cms_get     (CMSTREE *cms, const ITEM *items, ITEM n);
extern int      cms_super   (CMSTREE *cms,
                             const ITEM *items, ITEM n, RSUPP supp);
extern CMSTREE* cms_project (CMSTREE *dst, CMSTREE *src, ITEM item);
extern void     cms_prune   (CMSTREE *cms,
                             const ITEM *items, ITEM n, RSUPP supp);
extern int      cms_update  (CMSTREE *cms,
                             const ITEM *items, ITEM n, RSUPP supp);
extern int      cms_report  (CMSTREE *cms, ISREPORT *rep);

#ifndef NDEBUG
extern void     cms_show    (CMSTREE *cms, ITEMBASE *base, int ind);
#endif

/*----------------------------------------------------------------------
  Closed/Maximal Filter Functions
----------------------------------------------------------------------*/
extern CM4SEQS* cmx_create  (ITEM size, ITEM zmax, int target, int dir);
extern void     cmx_delete  (CM4SEQS *cmx);
extern ITEM     cmx_cnt     (CM4SEQS *cmx);
extern int      cmx_target  (CM4SEQS *cmx);
extern int      cmx_dir     (CM4SEQS *cmx);
extern RSUPP    cmx_supp    (CM4SEQS *cmx);
extern CMSTREE* cmx_tree    (CM4SEQS *cmx, ITEM i);

extern int      cmx_add     (CM4SEQS *cmx, ITEM item, RSUPP supp);
extern void     cmx_remove  (CM4SEQS *cmx, ITEM n);
extern int      cmx_update  (CM4SEQS *cmx,
                             const ITEM *items, ITEM n, RSUPP supp);
extern int      cmx_report  (CM4SEQS *cmx, ISREPORT *rep);
#ifndef NDEBUG
extern void     cmx_show    (CM4SEQS *cmx, ITEMBASE *base, int ind);
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define cms_create(s,t,d)   cms_createx(NULL, s, t, d)
#define cms_memsys(t)       ((t)->mem)
#define cms_cnt(t)          ((t)->cnt)
#define cms_dir(t)          ((t)->dir)
#define cms_supp(t)         ((t)->root.supp)
#define cms_target(t)       ((t)->target)
#define cms_nodecnt(t)      (ms_used((t)->mem))
#define cms_nodemax(t)      (ms_umax((t)->mem))

#define cmx_cnt(f)          ((f)->cnt)
#define cmx_target(f)       ((f)->target)
#define cmx_dir(f)          ((f)->dir)
#define cmx_tree(f,i)       ((f)->trees[i])
#define cmx_report(f,r)     cms_report((f)->trees[0], r)

#endif
